#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Initializer class for implementing Block Triangularization initialization
"""
from pyomo.environ import SolverFactory
from pyomo.common.config import Bool, ConfigDict, ConfigValue
from pyomo.contrib.incidence_analysis import (
    IncidenceGraphInterface,
    solve_strongly_connected_components,
)

from idaes.core.initialization.initializer_base import (
    InitializerBase,
    InitializationStatus,
)
from idaes.core.util.exceptions import InitializationError

__author__ = "Andrew Lee"


class BlockTriangularizationInitializer(InitializerBase):
    """
    Block Triangularization based Initializer object.

    This Initializer should be suitable for most models, but may struggle to initialize
    tightly coupled systems of equations.

    This Initializer uses the common workflow defined in InitializerBase and calls
    the Pyomo solve_strongly_connected_components function to initialize the model.

    """

    # TODO: Block solver is IPOPT for now, as fsolve struggles with VLE
    CONFIG = InitializerBase.CONFIG()
    CONFIG.declare(
        "block_solver",
        ConfigValue(
            default="ipopt_v2",
            description="Solver to use for NxN blocks",
        ),
    )
    CONFIG.declare(
        "block_solver_options",
        ConfigDict(
            implicit=True,
            description="Dict of options to pass to block solver",
            doc="Dict of options to use to set solver.options.",
        ),
    )
    CONFIG.block_solver_options.declare(
        "tol",
        ConfigValue(
            default=1e-8,
            domain=float,
            description="Convergence tolerance for block solver",
        ),
    )
    CONFIG.block_solver_options.declare(
        "max_iter",
        ConfigValue(
            default=200,
            domain=int,
            description="Iteration limit for block solver",
        ),
    )
    CONFIG.declare(
        "block_solver_writer_config",
        ConfigDict(
            implicit=True,
            description="Dict of writer_config arguments to pass to block solver",
        ),
    )
    CONFIG.block_solver_writer_config.declare(
        "linear_presolve",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Whether to use linear presolver with block solver",
        ),
    )
    CONFIG.block_solver_writer_config.declare(
        "scale_model",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Whether to apply model scaling with block solver",
        ),
    )
    CONFIG.declare(
        "block_solver_call_options",
        ConfigDict(
            implicit=True,
            description="Dict of arguments to pass to solver.solve call",
            doc="Dict of arguments to be passed as part of the solver.solve "
            "call, such as tee=True.",
        ),
    )
    CONFIG.declare(
        "calculate_variable_options",
        ConfigDict(
            implicit=True,
            description="Dict of options to pass to 1x1 block solver",
            doc="Dict of options to pass to calc_var_kwds argument in "
            "solve_strongly_connected_components method.",
        ),
    )

    def precheck(self, model):
        """
        Check for perfect matching in model.

        If this fails, it indicates a structural singularity in the model.
        """
        super().precheck(model)

        if model.is_indexed():
            for d in model.values():
                self._check_matching(d)
        else:
            self._check_matching(model)

    def _check_matching(self, block_data):
        """
        Run incidence analysis on given block data and check matching.
        """
        igraph = IncidenceGraphInterface(block_data, include_inequality=False)
        matching = igraph.maximum_matching()
        if len(matching) != len(igraph.variables):
            self._update_summary(
                block_data, "status", InitializationStatus.PrecheckFailed
            )
            raise InitializationError(
                f"Perfect matching not found for {block_data.name}. "
                f"This suggests that the model is structurally singular."
            )

    def initialization_routine(self, model):
        """
        Call Block Triangularization solver on model.
        """
        # TODO: For now, go directly through solver factory as default solver
        # options cause failures. Most of these appear to be due to scaling,
        # so hopefully we can fix these later.
        solver = SolverFactory(
            self.config.block_solver,
            options=self.config.block_solver_options,
            writer_config=self.config.block_solver_writer_config,
        )

        if model.is_indexed():
            for d in model.values():
                self._solve_block_data(d, solver)
        else:
            self._solve_block_data(model, solver)

    def _solve_block_data(self, block_data, solver):
        """
        Call solve_strongly_connected_components on a given BlockData.
        """
        # TODO: Can we get diagnostic output from this method?
        solve_strongly_connected_components(
            block_data,
            solver=solver,
            solve_kwds=self.config.block_solver_call_options,
            calc_var_kwds=self.config.calculate_variable_options,
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Property package for MEA Carbon Capture System Condenser

CO2-H2O VLE system
Ideal VLE with non-condensables
"""
# Import Python libraries
import logging

from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component, PhaseType as PT

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ideal import Ideal
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    IdealBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import fugacity

from idaes.models.properties.modular_properties.pure.Perrys import Perrys
from idaes.models.properties.modular_properties.pure.NIST import NIST

# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for an ideal CO2-H2O system
# Assumptions
# 1. Ideal gas-liquid system
# 2. CO2 conc. negligible in liquid phase, at equilibrium
# 3. Properties are applicable at standard pressure (1 atm)

# Reference Temperature : 298.15 K
# Reference Pressure : 101325 Pa

# Data Sources:

# [1] NIST Webbook, https://webbook.nist.gov/
#     Retrieved 27th November, 2020. Converted from bar to Pa
# [2] Perry's Chemical Engineers' Handbook 7th Ed.

# Initial temperature and pressure reference:
# Measurement and Modeling of the Phase Behavior of the
# (Carbon Dioxide + Water) Mixture at Temperatures From 298.15 K to 448.15
# KShu-Xin Hou, Geoffrey C. Maitland, and J. P. Martin Trusler*
# Qatar Carbonates and Carbon Storage Research Centre
# Department of Chemical Engineering Imperial College London,
# South Kensington Campus, London SW7 2AZ.  U.K

configuration = {
    # Specifying components
    "components": {
        "H2O": {
            "type": Component,
            "dens_mol_liq_comp": Perrys,
            "enth_mol_liq_comp": Perrys,
            "enth_mol_ig_comp": NIST,
            "pressure_sat_comp": NIST,
            "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
            "parameter_data": {
                "mw": (18.0153e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (220.64e5, pyunits.Pa),  # [1]
                "temperature_crit": (647, pyunits.K),  # [1]
                "dens_mol_liq_comp_coeff": {
                    "eqn_type": 1,
                    "1": (
                        5.459,
                        pyunits.kmol * pyunits.m**-3,
                    ),  # [2] pg. 2-98, temperature range 273.16 K - 333.15 K
                    "2": (0.30542, None),
                    "3": (647.13, pyunits.K),
                    "4": (0.081, None),
                },
                "cp_mol_ig_comp_coeff": {
                    "A": (
                        30.09200,
                        pyunits.J / pyunits.mol / pyunits.K,
                    ),  # [1] temperature range 500 K- 1700 K
                    "B": (
                        6.832514,
                        pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-1,
                    ),
                    "C": (
                        6.793435,
                        pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-2,
                    ),
                    "D": (
                        -2.534480,
                        pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-3,
                    ),
                    "E": (
                        0.082139,
                        pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**2,
                    ),
                    "F": (-250.8810, pyunits.kJ / pyunits.mol),
                    "G": (223.3967, pyunits.J / pyunits.mol / pyunits.K),
                    "H": (0, pyunits.kJ / pyunits.mol),
                },
                "cp_mol_liq_comp_coeff": {
                    "1": (
                        2.7637e5,
                        pyunits.J / pyunits.kmol / pyunits.K,
                    ),  # [2] pg 2-174, temperature range 273.16 K - 533.15 K
                    "2": (-2.0901e3, pyunits.J / pyunits.kmol / pyunits.K**2),
                    "3": (8.125, pyunits.J / pyunits.kmol / pyunits.K**3),
                    "4": (-1.4116e-2, pyunits.J / pyunits.kmol / pyunits.K**4),
                    "5": (9.3701e-6, pyunits.J / pyunits.kmol / pyunits.K**5),
                },
                "enth_mol_form_liq_comp_ref": (
                    -285.83e3,
                    pyunits.J / pyunits.mol,
                ),  # [1]
                "enth_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol),  # [1]
                "pressure_sat_comp_coeff": {
                    "A": (4.6543, None),  # [1], temperature range 255.9 K - 373 K
                    "B": (1435.264, pyunits.K),
                    "C": (-64.848, pyunits.K),
                },
            },
        },
        "CO2": {
            "type": Component,
            "valid_phase_types": PT.vaporPhase,
            "enth_mol_ig_comp": NIST,
            "parameter_data": {
                "mw": (44.0095e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (73.825e5, pyunits.Pa),  # [1]
                "temperature_crit": (304.23, pyunits.K),  # [1]
                "cp_mol_ig_comp_coeff": {  # [1], temperature range 298 K - 1200 K
                    "A": (24.99735, pyunits.J / pyunits.mol / pyunits.K),
                    "B": (
                        55.18696,
                        pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-1,
                    ),
                    "C": (
                        -33.69137,
                        pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-2,
                    ),
                    "D": (
                        7.948387,
                        pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-3,
                    ),
                    "E": (
                        -0.136638,
                        pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**2,
                    ),
                    "F": (-403.6075, pyunits.kJ / pyunits.mol),
                    "G": (228.2431, pyunits.J / pyunits.mol / pyunits.K),
                    "H": (0, pyunits.kJ / pyunits.mol),
                },
                "enth_mol_form_vap_comp_ref": (0, pyunits.J / pyunits.mol),  # [1]
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {"type": LiquidPhase, "equation_of_state": Ideal},
        "Vap": {"type": VaporPhase, "equation_of_state": Ideal},
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 10, 20, pyunits.mol / pyunits.s),
        "temperature": (273.15, 323.15, 1000, pyunits.K),
        "pressure": (5e4, 108900, 1e7, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": IdealBubbleDew,
}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pyomo.environ as pyo
from collections import deque
import pandas as pd
from idaes.apps.grid_integration import Tracker
from idaes.apps.grid_integration import Bidder
from idaes.apps.grid_integration import PlaceHolderForecaster
from idaes.apps.grid_integration.model_data import ThermalGeneratorModelData

from pyomo.common.dependencies import attempt_import

prescient_simulator, prescient_avail = attempt_import("prescient.simulator")


class ThermalGenerator:
    """
    Simple thermal generator model (MIP). Equations models are from Gao, Knueven,
    Siirola, Miller, Dowling (2022). Multiscale Simulation of Integrated Energy
    System and Electricity Market Interactions. Applied Energy.
    """

    # Using 4 segments to be consistent with models in RTS-GMLC dataset
    segment_number = 4

    def __init__(
        self,
        rts_gmlc_generator_dataframe,
        rts_gmlc_bus_dataframe,
        generator="102_STEAM_3",
    ):
        """
        Initializes the class object by building the thermal generator model.

        Arguments:
            rts_gmlc_generator_dataframe: the RTS-GMLC generator data in Pandas DataFrame
            horizon: the length of the planning horizon of the model.
            generator: a generator in RTS-GMLC

        Returns:
            None
        """

        self.generator = generator
        rts_gmlc_generator_dataframe = rts_gmlc_generator_dataframe.merge(
            rts_gmlc_bus_dataframe[["Bus ID", "Bus Name"]],
            how="left",
            left_on="Bus ID",
            right_on="Bus ID",
        )
        self._model_data_dict = self.assemble_model_data(
            generator_name=generator, gen_params=rts_gmlc_generator_dataframe
        )
        self.result_list = []

    def assemble_model_data(self, generator_name, gen_params):
        """
        This function assembles the parameter data to build the thermal generator
        model, given a list of generator names and the RTS-GMLC data directory.

        Arguments:
            generator_names: a generator name in RTS-GMLC dataset.
            gen_params: the RTS-GMLC generator data in Pandas DataFrame

        Returns:
            model_data: a dictionary which has this structure
            {data type name: value}.
        """

        gen_params = gen_params.set_index("GEN UID", inplace=False)
        properties = [
            "Bus Name",
            "PMin MW",
            "PMax MW",
            "Min Up Time Hr",
            "Min Down Time Hr",
            "Ramp Rate MW/Min",
            "Start Heat Warm MBTU",
            "Fuel Price $/MMBTU",
            "HR_avg_0",
            "HR_incr_1",
            "HR_incr_2",
            "HR_incr_3",
            "Output_pct_1",
            "Output_pct_2",
            "Output_pct_3",
        ]

        # to dict
        model_data = gen_params.loc[generator_name, properties].to_dict()

        model_data["RU"] = model_data["Ramp Rate MW/Min"] * 60
        model_data["RD"] = model_data["RU"]
        model_data["SU"] = min(model_data["PMin MW"], model_data["RU"])
        model_data["SD"] = min(model_data["PMin MW"], model_data["RD"])
        model_data["SU Cost"] = (
            model_data["Start Heat Warm MBTU"] * model_data["Fuel Price $/MMBTU"]
        )

        model_data["Min Load Cost"] = (
            model_data["HR_avg_0"]
            / 1000
            * model_data["Fuel Price $/MMBTU"]
            * model_data["PMin MW"]
        )

        model_data["Power Segments"] = {}
        model_data["Marginal Costs"] = {}

        model_data["Original Marginal Cost Curve"] = {}
        model_data["Original Marginal Cost Curve"][model_data["PMin MW"]] = (
            model_data["Min Load Cost"] / model_data["PMin MW"]
        )

        for l in range(1, ThermalGenerator.segment_number):
            model_data["Power Segments"][l] = (
                model_data["Output_pct_{}".format(l)] * model_data["PMax MW"]
            )
            model_data["Marginal Costs"][l] = (
                model_data["HR_incr_{}".format(l)]
                / 1000
                * model_data["Fuel Price $/MMBTU"]
            )
            model_data["Original Marginal Cost Curve"][
                model_data["Power Segments"][l]
            ] = model_data["Marginal Costs"][l]

        self._model_data = ThermalGeneratorModelData(
            gen_name=generator_name,
            bus=model_data["Bus Name"],
            p_min=model_data["PMin MW"],
            p_max=model_data["PMax MW"],
            min_down_time=model_data["Min Down Time Hr"],
            min_up_time=model_data["Min Up Time Hr"],
            ramp_up_60min=model_data["RU"],
            ramp_down_60min=model_data["RD"],
            shutdown_capacity=model_data["SD"],
            startup_capacity=model_data["SU"],
            initial_status=-1,
            initial_p_output=0,
            fixed_commitment=None,
            production_cost_bid_pairs=[
                (
                    model_data["PMin MW"],
                    model_data["Min Load Cost"] / model_data["PMin MW"],
                )
            ]
            + [
                (model_data["Power Segments"][l], model_data["Marginal Costs"][l])
                for l in range(1, ThermalGenerator.segment_number)
            ],
            startup_cost_pairs=[
                (model_data["Min Down Time Hr"], model_data["SU Cost"])
            ],
        )

        return model_data

    @property
    def model_data(self):
        return self._model_data

    @staticmethod
    def _add_UT_DT_constraints(b):
        """
        This function adds the minimum up/down time constraints using eq. 4 - 5
        in "On mixed-integer programming formulations for the unit commitment
        problem". INFORMS Journal on Computing, 32(4), pp.857-876. Knueven, B.,
        Ostrowski, J. and Watson, J.P., 2020.

        Arguments:
            b: a pyomo block

        Returns:
            None
        """

        def pre_shut_down_trajectory_set_rule(b):
            return (t for t in range(-pyo.value(b.min_dw_time) + 1, 0))

        b.pre_shut_down_trajectory_set = pyo.Set(
            dimen=1, initialize=pre_shut_down_trajectory_set_rule, ordered=True
        )

        def pre_start_up_trajectory_set_rule(b):
            return (t for t in range(-pyo.value(b.min_up_time) + 1, 0))

        b.pre_start_up_trajectory_set = pyo.Set(
            dimen=1, initialize=pre_start_up_trajectory_set_rule, ordered=True
        )

        b.pre_shut_down_trajectory = pyo.Param(
            b.pre_shut_down_trajectory_set, initialize=0, mutable=True
        )
        b.pre_start_up_trajectory = pyo.Param(
            b.pre_start_up_trajectory_set, initialize=0, mutable=True
        )

        def min_down_time_rule(b, h):
            if h < pyo.value(b.min_dw_time):
                return (
                    sum(
                        b.pre_shut_down_trajectory[h0]
                        for h0 in range(h - pyo.value(b.min_dw_time) + 1, 0)
                    )
                    + sum(b.shut_dw[h0] for h0 in range(h + 1))
                    <= 1 - b.on_off[h]
                )
            else:
                return (
                    sum(
                        b.shut_dw[h0]
                        for h0 in range(h - pyo.value(b.min_dw_time) + 1, h + 1)
                    )
                    <= 1 - b.on_off[h]
                )

        b.min_down_time_con = pyo.Constraint(b.HOUR, rule=min_down_time_rule)

        def min_up_time_rule(b, h):
            if h < pyo.value(b.min_up_time):
                return (
                    sum(
                        b.pre_start_up_trajectory[h0]
                        for h0 in range(h - pyo.value(b.min_up_time) + 1, 0)
                    )
                    + sum(b.start_up[h0] for h0 in range(h + 1))
                    <= b.on_off[h]
                )
            else:
                return (
                    sum(
                        b.start_up[h0]
                        for h0 in range(h - pyo.value(b.min_up_time) + 1, h + 1)
                    )
                    <= b.on_off[h]
                )

        b.min_up_time_con = pyo.Constraint(b.HOUR, rule=min_up_time_rule)

        return

    def populate_model(self, b, horizon):
        """
        This function builds the model for a thermal generator.

        Arguments:
            b: a pyomo block

        Returns:
            b: the constructed block.
        """

        model_data = self._model_data_dict

        ## define the sets
        b.HOUR = pyo.Set(initialize=list(range(horizon)))
        b.SEGMENTS = pyo.Set(initialize=list(range(1, ThermalGenerator.segment_number)))

        ## define the parameters

        b.start_up_cost = pyo.Param(initialize=model_data["SU Cost"], mutable=False)

        # capacity of generators: upper bound (MW)
        b.Pmax = pyo.Param(initialize=model_data["PMax MW"], mutable=False)

        # minimum power of generators: lower bound (MW)
        b.Pmin = pyo.Param(initialize=model_data["PMin MW"], mutable=False)

        b.power_segment_bounds = pyo.Param(
            b.SEGMENTS, initialize=model_data["Power Segments"], mutable=False
        )

        # get the cost slopes
        b.F = pyo.Param(
            b.SEGMENTS, initialize=model_data["Marginal Costs"], mutable=False
        )

        b.min_load_cost = pyo.Param(
            initialize=model_data["Min Load Cost"], mutable=False
        )

        # Ramp up limits (MW/h)
        b.ramp_up = pyo.Param(initialize=model_data["RU"], mutable=False)

        # Ramp down limits (MW/h)
        b.ramp_dw = pyo.Param(initialize=model_data["RD"], mutable=False)

        # start up ramp limit
        b.ramp_start_up = pyo.Param(initialize=model_data["SU"], mutable=False)

        # shut down ramp limit
        b.ramp_shut_dw = pyo.Param(initialize=model_data["SD"], mutable=False)

        # minimum down time [hr]
        b.min_dw_time = pyo.Param(
            initialize=int(model_data["Min Down Time Hr"]), mutable=False
        )

        # minimum up time [hr]
        b.min_up_time = pyo.Param(
            initialize=int(model_data["Min Up Time Hr"]), mutable=False
        )

        # on/off status from previous day
        b.pre_on_off = pyo.Param(within=pyo.Binary, default=1, mutable=True)

        # define a function to initialize the previous power params
        def init_pre_pow_fun(b):
            return b.pre_on_off * b.Pmin

        b.pre_P_T = pyo.Param(initialize=init_pre_pow_fun, mutable=True)

        ## define the variables

        # generator power (MW)

        # power generated by thermal generator
        b.P_T = pyo.Var(
            b.HOUR, initialize=model_data["PMin MW"], within=pyo.NonNegativeReals
        )

        # binary variables indicating on/off
        b.on_off = pyo.Var(b.HOUR, initialize=True, within=pyo.Binary)

        # binary variables indicating  start_up
        b.start_up = pyo.Var(b.HOUR, initialize=False, within=pyo.Binary)

        # binary variables indicating shut down
        b.shut_dw = pyo.Var(b.HOUR, initialize=False, within=pyo.Binary)

        # power produced in each segment
        b.power_segment = pyo.Var(
            b.HOUR, b.SEGMENTS, initialize=0, within=pyo.NonNegativeReals
        )

        ## Constraints

        # bounds on gen_pow
        def lhs_bnd_gen_pow_fun(b, h):
            return b.on_off[h] * b.Pmin <= b.P_T[h]

        b.lhs_bnd_gen_pow = pyo.Constraint(b.HOUR, rule=lhs_bnd_gen_pow_fun)

        def rhs_bnd_gen_pow_fun(b, h):
            return b.P_T[h] <= b.on_off[h] * b.Pmax

        b.rhs_bnd_gen_pow = pyo.Constraint(b.HOUR, rule=rhs_bnd_gen_pow_fun)

        # linearized power
        def linear_power_fun(b, h):
            return (
                b.P_T[h]
                == sum(b.power_segment[h, l] for l in b.SEGMENTS) + b.Pmin * b.on_off[h]
            )

        b.linear_power = pyo.Constraint(b.HOUR, rule=linear_power_fun)

        # bounds on segment power
        def seg_pow_bnd_fun(b, h, l):
            if l == 1:
                return (
                    b.power_segment[h, l]
                    <= (b.power_segment_bounds[l] - b.Pmin) * b.on_off[h]
                )
            else:
                return (
                    b.power_segment[h, l]
                    <= (b.power_segment_bounds[l] - b.power_segment_bounds[l - 1])
                    * b.on_off[h]
                )

        b.seg_pow_bnd = pyo.Constraint(b.HOUR, b.SEGMENTS, rule=seg_pow_bnd_fun)

        # start up and shut down logic (Garver 1962)
        def start_up_shut_dw_fun(b, h):
            if h == 0:
                return b.start_up[h] - b.shut_dw[h] == b.on_off[h] - b.pre_on_off
            else:
                return b.start_up[h] - b.shut_dw[h] == b.on_off[h] - b.on_off[h - 1]

        b.start_up_shut_dw = pyo.Constraint(b.HOUR, rule=start_up_shut_dw_fun)

        # either start up or shut down
        def start_up_or_shut_dw_fun(b, h):
            return b.start_up[h] + b.shut_dw[h] <= 1

        b.start_up_or_shut_dw = pyo.Constraint(b.HOUR, rule=start_up_or_shut_dw_fun)

        # ramp up limits
        def ramp_up_fun(b, h):
            """
            h stand for hour
            """
            if h == 0:
                return (
                    b.P_T[h]
                    <= b.pre_P_T
                    + b.ramp_up * b.pre_on_off
                    + b.ramp_start_up * b.start_up[h]
                )
            else:
                return (
                    b.P_T[h]
                    <= b.P_T[h - 1]
                    + b.ramp_up * b.on_off[h - 1]
                    + b.ramp_start_up * b.start_up[h]
                )

        b.ramp_up_con = pyo.Constraint(b.HOUR, rule=ramp_up_fun)

        # ramp shut down limits
        def ramp_shut_dw_fun(b, h):
            """
            h stand for hour.
            """
            if h == 0:
                return b.pre_P_T <= b.Pmax * b.on_off[h] + b.ramp_shut_dw * b.shut_dw[h]
            else:
                return (
                    b.P_T[h - 1] <= b.Pmax * b.on_off[h] + b.ramp_shut_dw * b.shut_dw[h]
                )

        b.ramp_shut_dw_con = pyo.Constraint(b.HOUR, rule=ramp_shut_dw_fun)

        # ramp down limits
        def ramp_dw_fun(b, h):
            """
            h stand for hour.
            """
            if h == 0:
                return (
                    b.pre_P_T - b.P_T[h]
                    <= b.ramp_dw * b.on_off[h] + b.ramp_shut_dw * b.shut_dw[h]
                )
            else:
                return (
                    b.P_T[h - 1] - b.P_T[h]
                    <= b.ramp_dw * b.on_off[h] + b.ramp_shut_dw * b.shut_dw[h]
                )

        b.ramp_dw_con = pyo.Constraint(b.HOUR, rule=ramp_dw_fun)

        ## add min up and down time constraints
        self._add_UT_DT_constraints(b)

        ## Expression
        def prod_cost_fun(b, h):
            return b.min_load_cost * b.on_off[h] + sum(
                b.F[l] * b.power_segment[h, l] for l in b.SEGMENTS
            )

        b.prod_cost_approx = pyo.Expression(b.HOUR, rule=prod_cost_fun)

        # start up costs
        def start_cost_fun(b, h):
            return b.start_up_cost * b.start_up[h]

        b.start_up_cost_expr = pyo.Expression(b.HOUR, rule=start_cost_fun)

        # total cost
        def tot_cost_fun(b, h):
            return b.prod_cost_approx[h] + b.start_up_cost_expr[h]

        b.tot_cost = pyo.Expression(b.HOUR, rule=tot_cost_fun)

        return

    @staticmethod
    def _update_UT_DT(b, implemented_shut_down, implemented_start_up):
        """
        This method updates the parameters in the minimum up/down time
        constraints based on the implemented shut down and start up events.

        Arguments:
            b: the block that needs to be updated
            implemented_shut_down: realized shut down events: [].
            implemented_start_up: realized start up events: []

        Returns:
            None
        """

        pre_shut_down_trajectory_copy = deque([])
        pre_start_up_trajectory_copy = deque([])

        # copy old trajectory
        for t in b.pre_shut_down_trajectory_set:
            pre_shut_down_trajectory_copy.append(
                round(pyo.value(b.pre_shut_down_trajectory[t]))
            )
        for t in b.pre_start_up_trajectory_set:
            pre_start_up_trajectory_copy.append(
                round(pyo.value(b.pre_start_up_trajectory[t]))
            )

        # add implemented trajectory to the queue
        pre_shut_down_trajectory_copy += deque(implemented_shut_down)
        pre_start_up_trajectory_copy += deque(implemented_start_up)

        # pop out outdated trajectory
        while len(pre_shut_down_trajectory_copy) > pyo.value(b.min_dw_time) - 1:
            pre_shut_down_trajectory_copy.popleft()
        while len(pre_start_up_trajectory_copy) > pyo.value(b.min_up_time) - 1:
            pre_start_up_trajectory_copy.popleft()

        # actual update
        for t in b.pre_shut_down_trajectory_set:
            b.pre_shut_down_trajectory[t] = pre_shut_down_trajectory_copy.popleft()

        for t in b.pre_start_up_trajectory_set:
            b.pre_start_up_trajectory[t] = pre_start_up_trajectory_copy.popleft()

        return

    @staticmethod
    def _update_power(b, implemented_power_output):
        """
        This method updates the parameters in the ramping constraints based on
        the implemented power outputs.

        Arguments:
            b: the block that needs to be updated
            implemented_power_output: realized power outputs: []

         Returns:
             None
        """

        b.pre_P_T = round(implemented_power_output[-1], 2)
        b.pre_on_off = round(int(implemented_power_output[-1] > 1e-3))

        return

    def update_model(
        self, b, implemented_shut_down, implemented_start_up, implemented_power_output
    ):
        """
        This method updates the parameters in the model based on
        the implemented power outputs, shut down and start up events.

        Arguments:
            b: the block that needs to be updated
            implemented_shut_down: realized shut down events: [].
            implemented_start_up: realized start up events: []
            implemented_power_output: realized power outputs: []

         Returns:
             None
        """

        self._update_UT_DT(b, implemented_shut_down, implemented_start_up)
        self._update_power(b, implemented_power_output)

        return

    @staticmethod
    def get_implemented_profile(b, last_implemented_time_step):
        """
        This method gets the implemented variable profiles in the last optimization
        solve.

        Arguments:
            b: the block
            model_var: intended variable name in str
            last_implemented_time_step: time index for the last implemented time
                                        step

         Returns:
             profile: the intended profile, {unit: [...]}
        """

        implemented_shut_down = deque(
            [pyo.value(b.shut_dw[t]) for t in range(last_implemented_time_step + 1)]
        )
        implemented_start_up = deque(
            [pyo.value(b.start_up[t]) for t in range(last_implemented_time_step + 1)]
        )
        implemented_power_output = deque(
            [pyo.value(b.P_T[t]) for t in range(last_implemented_time_step + 1)]
        )

        return {
            "implemented_shut_down": implemented_shut_down,
            "implemented_start_up": implemented_start_up,
            "implemented_power_output": implemented_power_output,
        }

    @staticmethod
    def get_last_delivered_power(b, last_implemented_time_step):
        """
        Returns the last delivered power output.

        Arguments:
            None

        Returns:
            None
        """

        return pyo.value(b.P_T[last_implemented_time_step])

    def record_results(self, b, date=None, hour=None, **kwargs):
        """
        Record the operations stats for the model.

        Arguments:

            date: current simulation date
            hour: current simulation hour

        Returns:
            None

        """

        df_list = []

        for t in b.HOUR:

            result_dict = {}
            result_dict["Generator"] = self.generator
            result_dict["Date"] = date
            result_dict["Hour"] = hour

            # simulation inputs
            result_dict["Horizon [hr]"] = int(t)

            # model vars
            result_dict["Thermal Power Generated [MW]"] = float(
                round(pyo.value(b.P_T[t]), 2)
            )

            result_dict["On/off [bin]"] = int(round(pyo.value(b.on_off[t])))
            result_dict["Start Up [bin]"] = int(round(pyo.value(b.start_up[t])))
            result_dict["Shut Down [bin]"] = int(round(pyo.value(b.shut_dw[t])))

            result_dict["Production Cost [$]"] = float(
                round(pyo.value(b.prod_cost_approx[t]), 2)
            )
            result_dict["Start-up Cost [$]"] = float(
                round(pyo.value(b.start_up_cost_expr[t]), 2)
            )
            result_dict["Total Cost [$]"] = float(round(pyo.value(b.tot_cost[t]), 2))

            # calculate mileage
            if t == 0:
                result_dict["Mileage [MW]"] = float(
                    round(abs(pyo.value(b.P_T[t] - b.pre_P_T)), 2)
                )
            else:
                result_dict["Mileage [MW]"] = float(
                    round(abs(pyo.value(b.P_T[t] - b.P_T[t - 1])), 2)
                )

            for key in kwargs:
                result_dict[key] = kwargs[key]

            result_df = pd.DataFrame.from_dict(result_dict, orient="index")
            df_list.append(result_df.T)

        # save the result to object property
        # wait to be written when simulation ends
        self.result_list.append(pd.concat(df_list))

        return

    def write_results(self, path):
        """
        This methods writes the saved operation stats into an csv file.

        Arguments:
            path: the path to write the results.

        Return:
            None
        """

        pd.concat(self.result_list).to_csv(path, index=False)

    @property
    def power_output(self):
        return "P_T"

    @property
    def total_cost(self):
        return ("tot_cost", 1)


if __name__ == "__main__":

    from idaes.apps.grid_integration.examples.utils import (
        rts_gmlc_generator_dataframe,
        rts_gmlc_bus_dataframe,
        prescient_5bus,
        daily_da_price_means,
        daily_rt_price_means,
        daily_da_price_stds,
        daily_rt_price_stds,
    )

    generator = "10_STEAM"
    horizon = 4
    solver = pyo.SolverFactory("cbc")

    run_tracker = True
    run_bidder = True
    run_prescient = True

    if run_tracker:

        # create a tracker model
        tracking_model_object = ThermalGenerator(
            rts_gmlc_generator_dataframe=rts_gmlc_generator_dataframe,
            rts_gmlc_bus_dataframe=rts_gmlc_bus_dataframe,
            generator=generator,
        )
        # make a tracker
        thermal_tracker = Tracker(
            tracking_model_object=tracking_model_object,
            tracking_horizon=horizon,
            n_tracking_hour=1,
            solver=solver,
        )

        market_dispatch = [30, 40, 50, 70]

        thermal_tracker.track_market_dispatch(
            market_dispatch=market_dispatch, date="2021-07-26", hour="17:00"
        )
        thermal_tracker.write_results(path="./")

    if run_bidder:

        # create a tracker model
        bidding_model_object = ThermalGenerator(
            rts_gmlc_generator_dataframe=rts_gmlc_generator_dataframe,
            rts_gmlc_bus_dataframe=rts_gmlc_bus_dataframe,
            generator=generator,
        )

        # create forecaster
        forecaster = PlaceHolderForecaster(
            daily_da_price_means=daily_da_price_means,
            daily_rt_price_means=daily_rt_price_means,
            daily_da_price_stds=daily_da_price_stds,
            daily_rt_price_stds=daily_rt_price_stds,
        )

        thermal_bidder = Bidder(
            bidding_model_object=bidding_model_object,
            day_ahead_horizon=48,
            real_time_horizon=4,
            n_scenario=10,
            solver=solver,
            forecaster=forecaster,
        )

        date = "2020-07-10"
        hour = 13
        bids = thermal_bidder.compute_day_ahead_bids(date, hour)
        thermal_bidder.write_results(path="./")

    if run_prescient and prescient_avail:

        options = {
            "data_path": prescient_5bus,
            "input_format": "rts-gmlc",
            "simulate_out_of_sample": True,
            "run_sced_with_persistent_forecast_errors": True,
            "output_directory": "bidding_plugin_test_thermal_generator",
            "start_date": "07-10-2020",
            "num_days": 2,
            "sced_horizon": 4,
            "compute_market_settlements": True,
            "day_ahead_pricing": "LMP",
            "ruc_mipgap": 0.01,
            "symbolic_solver_labels": True,
            "reserve_factor": 0.0,
            "deterministic_ruc_solver": "cbc",
            "sced_solver": "cbc",
            "plugin": {
                "doubleloop": {
                    "module": "thermal_generator_prescient_plugin.py",
                    "bidding_generator": generator,
                }
            },
        }

        prescient_simulator.Prescient().simulate(**options)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for 0D lumped capacitance heat exchanger model

Author: Rusty Gentile, John Eslick, Andrew Lee
"""
from sys import platform
import pytest
import re

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    TransformationFactory,
    value,
    Var,
    Param,
    units as pyunits,
)

from pyomo.dae import DerivativeVar
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import DynamicError, ConfigurationError, IdaesError

from idaes.core.util.testing import PhysicalParameterTestBlock, initialization_tester

from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)

from idaes.models.properties import iapws95
from idaes.models.unit_models import (
    HeatExchangerLumpedCapacitance,
    HeatExchangerFlowPattern,
)
from idaes.models.unit_models.heat_exchanger import HX0DInitializer
from idaes.models.unit_models.heat_exchanger import delta_temperature_lmtd_callback
from idaes.models.properties.general_helmholtz import helmholtz_available
from idaes.core.initialization import (
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox

# Get default solver for testing
solver = get_solver("ipopt_v2")

# Number of steps for transient simulations
TIME_STEPS = 50


class TestHXRegression(object):
    """
    Regression tests taken from the base 0D HeatExchanger model
    """

    @pytest.mark.unit
    def test_config(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = HeatExchangerLumpedCapacitance(
            hot_side_name="shell",
            cold_side_name="tube",
            shell={"property_package": m.fs.properties},
            tube={"property_package": m.fs.properties},
            dynamic_heat_balance=False,
        )

        # Check unit config arguments
        # There are 10 to 12 arguments since you can add a side 1 and 2 config by
        # side_1, side_2, or whatever the user named them, plus
        # "dynamic_heat_balance"
        assert len(m.fs.unit.config) >= 10 and len(m.fs.unit.config) <= 12

        assert not m.fs.unit.config.dynamic
        assert not m.fs.unit.config.has_holdup
        assert isinstance(m.fs.unit.config.shell, ConfigBlock)
        assert isinstance(m.fs.unit.config.tube, ConfigBlock)
        assert (
            m.fs.unit.config.delta_temperature_callback
            is delta_temperature_lmtd_callback
        )
        assert m.fs.unit.config.flow_pattern == HeatExchangerFlowPattern.countercurrent

        # Check shell config
        assert len(m.fs.unit.config.shell) == 7
        assert (
            m.fs.unit.config.shell.material_balance_type
            == MaterialBalanceType.useDefault
        )
        assert (
            m.fs.unit.config.shell.energy_balance_type == EnergyBalanceType.useDefault
        )
        assert (
            m.fs.unit.config.shell.momentum_balance_type
            == MomentumBalanceType.pressureTotal
        )
        assert not m.fs.unit.config.shell.has_phase_equilibrium
        assert not m.fs.unit.config.shell.has_pressure_change
        assert m.fs.unit.config.shell.property_package is m.fs.properties

        # Check tube config
        assert len(m.fs.unit.config.tube) == 7
        assert (
            m.fs.unit.config.tube.material_balance_type
            == MaterialBalanceType.useDefault
        )
        assert m.fs.unit.config.tube.energy_balance_type == EnergyBalanceType.useDefault
        assert (
            m.fs.unit.config.tube.momentum_balance_type
            == MomentumBalanceType.pressureTotal
        )
        assert not m.fs.unit.config.tube.has_phase_equilibrium
        assert not m.fs.unit.config.tube.has_pressure_change
        assert m.fs.unit.config.tube.property_package is m.fs.properties

        assert m.fs.unit.default_initializer is HX0DInitializer

    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = HeatExchangerLumpedCapacitance(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            flow_pattern=HeatExchangerFlowPattern.crossflow,
            dynamic_heat_balance=False,
        )

        m.fs.unit.hot_side_inlet.flow_vol[0].fix(1e-3)
        m.fs.unit.hot_side_inlet.temperature[0].fix(320)
        m.fs.unit.hot_side_inlet.pressure[0].fix(101325)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        m.fs.unit.cold_side_inlet.flow_vol[0].fix(1e-3)
        m.fs.unit.cold_side_inlet.temperature[0].fix(300)
        m.fs.unit.cold_side_inlet.pressure[0].fix(101325)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

        m.fs.unit.area.fix(1000)

        # Modified from the original:
        # m.fs.unit.overall_heat_transfer_coefficient.fix(100)
        m.fs.unit.ua_hot_side.fix(200 * 1000)
        m.fs.unit.ua_cold_side.fix(200 * 1000)

        m.fs.unit.crossflow_factor.fix(0.6)

        return m

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_sapon_solution(self, sapon):
        results = solver.solve(sapon)
        check_optimal_termination(results)

        assert pytest.approx(1e-3, abs=1e-6) == value(
            sapon.fs.unit.hot_side_outlet.flow_vol[0]
        )
        assert pytest.approx(1e-3, abs=1e-6) == value(
            sapon.fs.unit.cold_side_outlet.flow_vol[0]
        )

        assert pytest.approx(55388.0, rel=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "H2O"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "Ethanol"]
        )

        assert pytest.approx(55388.0, rel=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "H2O"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "Ethanol"]
        )

        assert pytest.approx(301.3, abs=1e-1) == value(
            sapon.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(318.7, abs=1e-1) == value(
            sapon.fs.unit.cold_side_outlet.temperature[0]
        )

        assert pytest.approx(101325, abs=1e2) == value(
            sapon.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(101325, abs=1e2) == value(
            sapon.fs.unit.cold_side_outlet.pressure[0]
        )


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestHXLCGeneric(object):
    @pytest.fixture()
    def static_flowsheet_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = iapws95.Iapws95ParameterBlock()
        return m

    @pytest.fixture()
    def dynamic_flowsheet_model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 1], time_units=pyunits.s)
        m.fs.properties = iapws95.Iapws95ParameterBlock()
        return m

    @pytest.fixture()
    def unconstrained_model(self, dynamic_flowsheet_model):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 1], time_units=pyunits.s)
        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = HeatExchangerLumpedCapacitance(
            hot_side_name="shell",
            cold_side_name="tube",
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            dynamic=False,
        )

        return m

    @pytest.fixture
    def model(self, unconstrained_model):
        m = unconstrained_model
        m.discretizer = TransformationFactory("dae.finite_difference")
        m.discretizer.apply_to(m, nfe=1, wrt=m.fs.time, scheme="BACKWARD")

        m.fs.unit.area.fix(1000)
        m.fs.unit.heat_capacity = 100

        m.fs.unit.shell_inlet.flow_mol[:].fix(100)
        m.fs.unit.shell_inlet.pressure[:].fix(101325)
        m.fs.unit.shell_inlet.enth_mol[:].fix(4000)

        m.fs.unit.tube_inlet.flow_mol[:].fix(100)
        m.fs.unit.tube_inlet.pressure[:].fix(101325)
        m.fs.unit.tube_inlet.enth_mol[:].fix(3000)

        m.fs.unit.ua_cold_side[:].fix(100)
        m.fs.unit.ua_hot_side[:].fix(100)

        return m

    @pytest.mark.component
    def test_structural_issues(self, model):
        dt = DiagnosticsToolbox(model)
        dt.display_potential_evaluation_errors()
        # TODO: Evaluation errors due to temperature differentials
        # TODO: Skip unit consistency due to Pyomo DAE issue
        dt.assert_no_structural_warnings(
            ignore_evaluation_errors=True, ignore_unit_consistency=True
        )

    @pytest.mark.unit
    def test_units_unconstrained(self, unconstrained_model):
        # ...but without the discretizer, the units are fine
        assert_units_consistent(unconstrained_model)

        m = unconstrained_model
        assert_units_equivalent(m.fs.unit.ua_hot_side, pyunits.W / pyunits.K)
        assert_units_equivalent(m.fs.unit.ua_cold_side, pyunits.W / pyunits.K)
        assert_units_equivalent(m.fs.unit.ua_hot_side_to_wall, pyunits.W / pyunits.K)
        assert_units_equivalent(m.fs.unit.temperature_wall, pyunits.K)
        assert_units_equivalent(m.fs.unit.heat_capacity_wall, pyunits.J / pyunits.K)
        assert_units_equivalent(
            m.fs.unit.thermal_resistance_wall, pyunits.K / pyunits.W
        )
        assert_units_equivalent(
            m.fs.unit.thermal_fouling_hot_side, pyunits.K / pyunits.W
        )
        assert_units_equivalent(
            m.fs.unit.thermal_fouling_cold_side, pyunits.K / pyunits.W
        )
        assert_units_equivalent(m.fs.unit.dT_wall_dt, pyunits.K / pyunits.s)

    @pytest.mark.unit
    def test_base_vars(self, model):
        assert isinstance(model.fs.unit.area, Var)
        assert isinstance(model.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(model.fs.unit.heat_duty, Var)
        assert isinstance(model.fs.unit.delta_temperature_in, Var)
        assert isinstance(model.fs.unit.delta_temperature_out, Var)

    @pytest.mark.unit
    def test_new_vars(self, model):
        assert isinstance(model.fs.unit.dT_wall_dt, DerivativeVar)
        assert isinstance(model.fs.unit.temperature_wall, Var)
        assert isinstance(model.fs.unit.ua_cold_side, Var)
        assert isinstance(model.fs.unit.ua_hot_side, Var)
        assert isinstance(model.fs.unit.ua_hot_side_to_wall, Var)
        assert isinstance(model.fs.unit.heat_capacity_wall, Param)
        assert isinstance(model.fs.unit.thermal_resistance_wall, Param)
        assert isinstance(model.fs.unit.thermal_fouling_hot_side, Param)
        assert isinstance(model.fs.unit.thermal_fouling_cold_side, Param)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        model.fs.unit.initialize()
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

        # Combine with solve test due to how fixtures are set up
        dt = DiagnosticsToolbox(model)
        dt.display_constraints_with_large_residuals()
        dt.assert_no_numerical_warnings()

    @pytest.mark.unit
    def test_dynamic_heat_in_static_flowsheet(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(
            DynamicError,
            match="trying to declare a dynamic model within "
            "a steady-state flowsheet. This is not "
            "supported by the IDAES framework. Try "
            "creating a dynamic flowsheet instead, and "
            "declaring some models as steady-state.",
        ):
            m.fs.unit = HeatExchangerLumpedCapacitance(
                hot_side_name="shell",
                cold_side_name="tube",
                shell={"property_package": m.fs.properties},
                tube={"property_package": m.fs.properties},
                dynamic=False,
                dynamic_heat_balance=True,
            )

    @pytest.mark.unit
    def test_dynamic_cv_wo_dynamic_heat(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(
            ConfigurationError,
            match="dynamic can only be True if dynamic_heat_balance " "is also True.",
        ):
            m.fs.unit = HeatExchangerLumpedCapacitance(
                hot_side_name="shell",
                cold_side_name="tube",
                shell={"property_package": m.fs.properties},
                tube={"property_package": m.fs.properties},
                dynamic=True,
                dynamic_heat_balance=False,
            )

    @pytest.mark.unit
    def test_default_in_dynamic_flowsheet(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 1], time_units=pyunits.s)

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = HeatExchangerLumpedCapacitance(
            hot_side_name="shell",
            cold_side_name="tube",
            shell={"property_package": m.fs.properties},
            tube={"property_package": m.fs.properties},
        )

        assert m.fs.unit.config.dynamic_heat_balance
        assert m.fs.unit.config.dynamic
        assert m.fs.unit.config.has_holdup

    @pytest.mark.unit
    def test_steady_state_cv_in_dynamic_flowsheet(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 1], time_units=pyunits.s)

        m.fs.properties = PhysicalParameterTestBlock()

        m.fs.unit = HeatExchangerLumpedCapacitance(
            hot_side_name="shell",
            cold_side_name="tube",
            shell={"property_package": m.fs.properties},
            tube={"property_package": m.fs.properties},
            dynamic=False,
            dynamic_heat_balance=True,
        )

        assert m.fs.unit.config.dynamic_heat_balance
        assert not m.fs.unit.config.dynamic
        assert not m.fs.unit.config.has_holdup

    @pytest.mark.unit
    def test_dynamic_wo_holdup(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=True, time_set=[0, 1], time_units=pyunits.s)

        m.fs.properties = PhysicalParameterTestBlock()

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "invalid arguments for dynamic and has_holdup. "
                "If dynamic = True, has_holdup must also be True "
                "(was False)"
            ),
        ):
            m.fs.unit = HeatExchangerLumpedCapacitance(
                hot_side_name="shell",
                cold_side_name="tube",
                shell={"property_package": m.fs.properties},
                tube={"property_package": m.fs.properties},
                dynamic=True,
                dynamic_heat_balance=True,
                has_holdup=False,
            )

    @pytest.mark.unit
    def test_static_heat_balance(self, static_flowsheet_model):
        m = static_flowsheet_model
        m.fs.unit = HeatExchangerLumpedCapacitance(
            hot_side_name="shell",
            cold_side_name="tube",
            shell={"property_package": m.fs.properties},
            tube={"property_package": m.fs.properties},
            dynamic=False,
            dynamic_heat_balance=False,
        )

        with pytest.raises(
            IdaesError,
            match="heat holdup term cannot be activated when "
            "`dynamic_heat_balance=False`",
        ):
            m.fs.unit.activate_dynamic_heat_eq()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, model):
        perf_dict = model.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "HX Area": model.fs.unit.area,
                "Heat Duty": model.fs.unit.heat_duty[0],
                "HX Coefficient": model.fs.unit.overall_heat_transfer_coefficient[0],
                "Delta T In": model.fs.unit.delta_temperature_in[0],
                "Delta T Out": model.fs.unit.delta_temperature_out[0],
            },
            "exprs": {
                "Delta T Driving": model.fs.unit.delta_temperature[0],
            },
        }


class TestInitializers:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = HeatExchangerLumpedCapacitance(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            flow_pattern=HeatExchangerFlowPattern.crossflow,
            dynamic_heat_balance=False,
        )

        m.fs.unit.hot_side_inlet.flow_vol[0].set_value(1e-3)
        m.fs.unit.hot_side_inlet.temperature[0].set_value(320)
        m.fs.unit.hot_side_inlet.pressure[0].set_value(101325)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "H2O"].set_value(55388.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "NaOH"].set_value(100.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "EthylAcetate"].set_value(100.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "SodiumAcetate"].set_value(0.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "Ethanol"].set_value(0.0)

        m.fs.unit.cold_side_inlet.flow_vol[0].set_value(1e-3)
        m.fs.unit.cold_side_inlet.temperature[0].set_value(300)
        m.fs.unit.cold_side_inlet.pressure[0].set_value(101325)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "H2O"].set_value(55388.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "NaOH"].set_value(100.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "EthylAcetate"].set_value(100.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "SodiumAcetate"].set_value(0.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "Ethanol"].set_value(0.0)

        m.fs.unit.area.fix(1000)

        # Modified from the original:
        m.fs.unit.ua_hot_side.fix(200 * 1000)
        m.fs.unit.ua_cold_side.fix(200 * 1000)

        m.fs.unit.crossflow_factor.fix(0.6)

        return m

    @pytest.mark.component
    def test_hx_initializer(self, model):
        # Setting bounds on deltaT seems to help avoid a temperature cross-over
        # during initialization.
        model.fs.unit.delta_temperature_in.setlb(1e-8)
        model.fs.unit.delta_temperature_out.setlb(1e-8)
        # TODO: Linear presolve appears to cause issues on Windows
        # Hope that these will be fixed with better scaling
        initializer = HX0DInitializer(writer_config={"linear_presolve": False})
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(1e-3, abs=1e-6) == value(
            model.fs.unit.hot_side_outlet.flow_vol[0]
        )
        assert pytest.approx(1e-3, abs=1e-6) == value(
            model.fs.unit.cold_side_outlet.flow_vol[0]
        )

        assert pytest.approx(55388.0, rel=1e-3) == value(
            model.fs.unit.hot_side_outlet.conc_mol_comp[0, "H2O"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            model.fs.unit.hot_side_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            model.fs.unit.hot_side_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            model.fs.unit.hot_side_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            model.fs.unit.hot_side_outlet.conc_mol_comp[0, "Ethanol"]
        )

        assert pytest.approx(55388.0, rel=1e-3) == value(
            model.fs.unit.cold_side_outlet.conc_mol_comp[0, "H2O"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            model.fs.unit.cold_side_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            model.fs.unit.cold_side_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            model.fs.unit.cold_side_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            model.fs.unit.cold_side_outlet.conc_mol_comp[0, "Ethanol"]
        )

        assert pytest.approx(301.3, abs=1e-1) == value(
            model.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(318.7, abs=1e-1) == value(
            model.fs.unit.cold_side_outlet.temperature[0]
        )

        assert pytest.approx(101325, abs=1e2) == value(
            model.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(101325, abs=1e2) == value(
            model.fs.unit.cold_side_outlet.pressure[0]
        )

        assert not model.fs.unit.hot_side_inlet.flow_vol[0].fixed
        assert not model.fs.unit.hot_side_inlet.temperature[0].fixed
        assert not model.fs.unit.hot_side_inlet.pressure[0].fixed
        assert not model.fs.unit.hot_side_inlet.conc_mol_comp[0, "H2O"].fixed
        assert not model.fs.unit.hot_side_inlet.conc_mol_comp[0, "NaOH"].fixed
        assert not model.fs.unit.hot_side_inlet.conc_mol_comp[0, "EthylAcetate"].fixed
        assert not model.fs.unit.hot_side_inlet.conc_mol_comp[0, "SodiumAcetate"].fixed
        assert not model.fs.unit.hot_side_inlet.conc_mol_comp[0, "Ethanol"].fixed

        assert not model.fs.unit.cold_side_inlet.flow_vol[0].fixed
        assert not model.fs.unit.cold_side_inlet.temperature[0].fixed
        assert not model.fs.unit.cold_side_inlet.pressure[0].fixed
        assert not model.fs.unit.cold_side_inlet.conc_mol_comp[0, "H2O"].fixed
        assert not model.fs.unit.cold_side_inlet.conc_mol_comp[0, "NaOH"].fixed
        assert not model.fs.unit.cold_side_inlet.conc_mol_comp[0, "EthylAcetate"].fixed
        assert not model.fs.unit.cold_side_inlet.conc_mol_comp[0, "SodiumAcetate"].fixed
        assert not model.fs.unit.cold_side_inlet.conc_mol_comp[0, "Ethanol"].fixed

    # TODO: BT initializer runs into a Max Iter failure on Windows



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This provides standard valve models for adiabatic control valves.  Beyond the
most common valve models, and adiabatic valve model can be added by supplying
custom callbacks for the pressure-flow relation or valve function.
"""
# Changing existing config block attributes
# pylint: disable=protected-access

__Author__ = "John Eslick"

from enum import Enum

import pyomo.environ as pyo
from pyomo.common.config import ConfigValue, In

from idaes.core import declare_process_block_class
from idaes.models.unit_models.pressure_changer import (
    PressureChangerData,
    ThermodynamicAssumption,
    MaterialBalanceType,
)
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

_log = idaeslog.getLogger(__name__)


class ValveFunctionType(Enum):
    """
    Enum of supported valve types.
    """

    linear = 1
    quick_opening = 2
    equal_percentage = 3


def linear_cb(valve):
    """
    Linear opening valve function callback.
    """

    @valve.Expression(valve.flowsheet().time)
    def valve_function(b, t):
        return b.valve_opening[t]


def quick_cb(valve):
    """
    Quick opening valve function callback.
    """

    @valve.Expression(valve.flowsheet().time)
    def valve_function(b, t):
        return pyo.sqrt(b.valve_opening[t])


def equal_percentage_cb(valve):
    """
    Equal percentage valve function callback.
    """
    valve.alpha = pyo.Var(initialize=100, doc="Valve function parameter")
    valve.alpha.fix()

    @valve.Expression(valve.flowsheet().time)
    def valve_function(b, t):
        return b.alpha ** (b.valve_opening[t] - 1)


def pressure_flow_default_callback(valve):
    """
    Add the default pressure flow relation constraint.  This will be used in the
    valve model, a custom callback is provided.
    """
    umeta = (
        valve.control_volume.config.property_package.get_metadata().get_derived_units
    )

    valve.Cv = pyo.Var(
        initialize=0.1,
        doc="Valve flow coefficient",
        units=umeta("amount") / umeta("time") / umeta("pressure") ** 0.5,
    )
    valve.Cv.fix()

    valve.flow_var = pyo.Reference(valve.control_volume.properties_in[:].flow_mol)
    valve.pressure_flow_equation_scale = lambda x: x**2

    @valve.Constraint(valve.flowsheet().time)
    def pressure_flow_equation(b, t):
        Po = b.control_volume.properties_out[t].pressure
        Pi = b.control_volume.properties_in[t].pressure
        F = b.control_volume.properties_in[t].flow_mol
        Cv = b.Cv
        fun = b.valve_function[t]
        return F**2 == Cv**2 * (Pi - Po) * fun**2


@declare_process_block_class("Valve", doc="Adiabatic valves")
class ValveData(PressureChangerData):
    """
    Basic valve model class.
    """

    # Same settings as the default pressure changer, but force to expander with
    # isentropic efficiency
    CONFIG = PressureChangerData.CONFIG()
    CONFIG.compressor = False
    CONFIG.get("compressor")._default = False
    CONFIG.get("compressor")._domain = In([False])
    CONFIG.material_balance_type = MaterialBalanceType.componentTotal
    CONFIG.get("material_balance_type")._default = MaterialBalanceType.componentTotal
    CONFIG.thermodynamic_assumption = ThermodynamicAssumption.adiabatic
    CONFIG.get("thermodynamic_assumption")._default = ThermodynamicAssumption.adiabatic
    CONFIG.get("thermodynamic_assumption")._domain = In(
        [ThermodynamicAssumption.adiabatic]
    )
    CONFIG.declare(
        "valve_function_callback",
        ConfigValue(
            default=ValveFunctionType.linear,
            description="Valve function type or callback for custom",
            doc="""This takes either an enumerated valve function type in: {
ValveFunctionType.linear, ValveFunctionType.quick_opening,
ValveFunctionType.equal_percentage, ValveFunctionType.custom} or a callback
function that takes a valve model object as an argument and adds a time-indexed
valve_function expression to it. Any additional required variables, expressions,
or constraints required can also be added by the callback.""",
        ),
    )
    CONFIG.declare(
        "pressure_flow_callback",
        ConfigValue(
            default=pressure_flow_default_callback,
            description="Callback function providing the valve_function expression",
            doc="""This callback function takes a valve model object as an argument
and adds a time-indexed valve_function expression to it.  Any additional required
variables, expressions, or constraints required can also be added by the callback.""",
        ),
    )

    def build(self):
        super().build()

        self.valve_opening = pyo.Var(
            self.flowsheet().time,
            initialize=1,
            bounds=(0, 1),
            doc="Fraction open for valve from 0 to 1",
        )
        self.valve_opening.fix()
        # If the valve function callback is set to one of the known enumerated
        # types, set the callback appropriately.  If not callable and not a known
        # type raise ConfigurationError.
        vfcb = self.config.valve_function_callback
        if not callable(vfcb):
            if vfcb == ValveFunctionType.linear:
                self.config.valve_function_callback = linear_cb
            elif vfcb == ValveFunctionType.quick_opening:
                self.config.valve_function_callback = quick_cb
            elif vfcb == ValveFunctionType.equal_percentage:
                self.config.valve_function_callback = equal_percentage_cb
            else:
                raise ConfigurationError("Invalid valve function callback.")
        self.config.valve_function_callback(self)
        self.config.pressure_flow_callback(self)

    def initialize_build(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialize the valve based on a deltaP guess.

        Args:
            state_args (dict): Initial state for property initialization
            outlvl : sets output level of initialization routine
            solver (str): Solver to use for initialization
            optarg (dict): Solver arguments dictionary
        """

        for t in self.flowsheet().time:
            if (
                self.deltaP[t].fixed
                or self.ratioP[t].fixed
                or self.outlet.pressure[t].fixed
            ):
                continue
            # Generally for the valve initialization pressure drop won't be
            # fixed, so if there is no good guess on deltaP try to out one in
            Pout = self.outlet.pressure[t]
            Pin = self.inlet.pressure[t]
            if self.deltaP[t].value is not None:
                prdp = pyo.value((self.deltaP[t] - Pin) / Pin)
            else:
                prdp = -100  # crazy number to say don't use deltaP as guess
            if pyo.value(Pout / Pin) > 1 or pyo.value(Pout / Pin) < 0.0:
                if pyo.value(self.ratioP[t]) <= 1 and pyo.value(self.ratioP[t]) >= 0:
                    Pout.value = pyo.value(Pin * self.ratioP[t])
                elif prdp <= 1 and prdp >= 0:
                    Pout.value = pyo.value(prdp * Pin)
                else:
                    Pout.value = pyo.value(Pin * 0.95)
            self.deltaP[t] = pyo.value(Pout - Pin)
            self.ratioP[t] = pyo.value(Pout / Pin)

        # one bad thing about reusing this is that the log messages aren't
        # really compatible with being nested inside another initialization
        super().initialize_build(
            state_args=state_args, outlvl=outlvl, solver=solver, optarg=optarg
        )

    def calculate_scaling_factors(self):
        """
        Calculate pressure flow constraint scaling from flow variable scale.
        """
        # The value of the valve opening and the output of the valve function
        # expression are between 0 and 1, so the only thing that needs to be
        # scaled here is the pressure-flow constraint, which can be scaled by
        # using the flow variable scale.  The flow variable could be defined
        # in different ways, so the flow variable is determined here from a
        # "flow_var[t]" reference set in the pressure-flow callback. The flow
        # term could be in various forms, so an optional
        # "pressure_flow_equation_scale" function can be defined in the callback
        # as well.  The pressure-flow function could be flow = f(Pin, Pout), but
        # it could also be flow**2 = f(Pin, Pout), ... The so
        # "pressure_flow_equation_scale" provides the form of the LHS side as
        # a function of the flow variable.

        super().calculate_scaling_factors()

        # Do some error trapping.
        if not hasattr(self, "pressure_flow_equation"):
            raise AttributeError(
                "Pressure-flow callback must define pressure_flow_equation[t]"
            )
        # Check for flow term form if none assume flow = f(Pin, Pout)
        if hasattr(self, "pressure_flow_equation_scale"):
            ff = self.pressure_flow_equation_scale
        else:
            # pylint: disable-next=unnecessary-lambda-assignment
            ff = lambda x: x
        # if the "flow_var" is not set raise an exception
        if not hasattr(self, "flow_var"):
            raise AttributeError(
                "Pressure-flow callback must define flow_var[t] reference"
            )

        # Calculate and set the pressure-flow relation scale.
        if hasattr(self, "pressure_flow_equation"):
            for t, c in self.pressure_flow_equation.items():
                iscale.constraint_scaling_transform(
                    c,
                    ff(
                        iscale.get_scaling_factor(
                            self.flow_var[t], default=1, warning=True
                        )
                    ),
                )

    def _get_performance_contents(self, time_point=0):
        pc = super()._get_performance_contents(time_point=time_point)

        pc["vars"]["Opening"] = self.valve_opening[time_point]
        try:
            pc["vars"]["Valve Coefficient"] = self.Cv
        except AttributeError:
            pass
        if self.config.valve_function_callback == ValveFunctionType.equal_percentage:
            pc["vars"]["alpha"] = self.alpha
        return pc



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring

from .base.process_base import ProcessBlockData, useDefault, MaterialFlowBasis
from .base.process_block import ProcessBlock, declare_process_block_class
from .base.unit_model import UnitModelBlockData, UnitModelBlock
from .base.flowsheet_model import FlowsheetBlockData, FlowsheetBlock
from .base.property_base import StateBlockData, PhysicalParameterBlock, StateBlock
from .base.reaction_base import (
    ReactionBlockDataBase,
    ReactionParameterBlock,
    ReactionBlockBase,
)
from .base.control_volume_base import (
    ControlVolumeBlockData,
    CONFIG_Template,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
)
from .base.control_volume0d import ControlVolume0DBlock
from .base.control_volume1d import ControlVolume1DBlock, DistributedVars
from .base.phases import (
    Phase,
    LiquidPhase,
    SolidPhase,
    VaporPhase,
    PhaseType,
    AqueousPhase,
)
from .base.components import Component, Solvent, Solute, Ion, Cation, Anion, Apparent
from .base.costing_base import (
    FlowsheetCostingBlock,
    FlowsheetCostingBlockData,
    UnitModelCostingBlock,
    register_idaes_currency_units,
)
from .base.var_like_expression import VarLikeExpression
from .base.property_set import StandardPropertySet, ElectrolytePropertySet



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Unit operation model for a waterpipe with pressure drop:

Created on Aug 27, 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models_extra.power_generation.unit_models.waterpipe import WaterPipe

# Import Unit Model Modules
from idaes.models.properties import iapws95

from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def build_unit():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterPipe(
        dynamic=False,
        property_package=m.fs.properties,
        has_holdup=True,
        has_heat_transfer=False,
        has_pressure_change=True,
        water_phase="Liq",
        contraction_expansion_at_end="contraction",
    )

    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(build_unit):
    """Make a unit model and make sure it doesn't throw exception"""
    m = build_unit
    assert degrees_of_freedom(m) == 9
    # Check unit config arguments
    assert len(m.fs.unit.config) == 11
    assert m.fs.unit.config.has_pressure_change
    assert isinstance(m.fs.unit.deltaP, pyo.Var)
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unit(build_unit):
    m = build_unit
    m.fs.unit.diameter.fix(0.04)
    m.fs.unit.length.fix(40)
    m.fs.unit.number_of_pipes.fix(100)
    m.fs.unit.elevation_change.fix(25)
    m.fs.unit.area_ratio.fix(0.5)
    m.fs.unit.fcorrection_dp.fix(1.0)

    state_args = {"flow_mol": 10000, "pressure": 1.3e7, "enth_mol": 18000}

    initialization_tester(build_unit, dof=3, state_args=state_args)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unit(build_unit):
    m = build_unit

    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)

    # check material balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].flow_mol
                - m.fs.unit.control_volume.properties_out[0].flow_mol
            ),
            abs=1e-3,
        )
        == 0
    )

    # pressure drop
    assert pytest.approx(-224074.4039, rel=1e-5) == pyo.value(m.fs.unit.deltaP[0])
    # check energy balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].enth_mol
                - m.fs.unit.control_volume.properties_out[0].enth_mol
            ),
            abs=1e-3,
        )
        == 0
    )


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_pipe_expansion():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterPipe(
        dynamic=False,
        property_package=m.fs.properties,
        has_holdup=True,
        has_heat_transfer=False,
        has_pressure_change=True,
        water_phase="Liq",
        contraction_expansion_at_end="expansion",
    )
    m.fs.unit.diameter.fix(0.04)
    m.fs.unit.length.fix(40)
    m.fs.unit.number_of_pipes.fix(100)
    m.fs.unit.elevation_change.fix(25)
    m.fs.unit.area_ratio.fix(0.5)
    m.fs.unit.fcorrection_dp.fix(1.0)

    state_args = {"flow_mol": 10000, "pressure": 1.3e7, "enth_mol": 18000}

    initialization_tester(m, dof=3, state_args=state_args)
    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)

    # check material balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].flow_mol
                - m.fs.unit.control_volume.properties_out[0].flow_mol
            ),
            abs=1e-3,
        )
        == 0
    )

    # pressure drop
    assert pytest.approx(-224306.548, rel=1e-5) == pyo.value(m.fs.unit.deltaP[0])
    # check energy balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].enth_mol
                - m.fs.unit.control_volume.properties_out[0].enth_mol
            ),
            abs=1e-3,
        )
        == 0
    )


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_pipe_noexpansion():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterPipe(
        dynamic=False,
        property_package=m.fs.properties,
        has_holdup=True,
        has_heat_transfer=False,
        has_pressure_change=True,
        water_phase="Liq",
        contraction_expansion_at_end="None",
    )
    m.fs.unit.diameter.fix(0.04)
    m.fs.unit.length.fix(40)
    m.fs.unit.number_of_pipes.fix(100)
    m.fs.unit.elevation_change.fix(25)
    m.fs.unit.fcorrection_dp.fix(1.0)

    state_args = {"flow_mol": 10000, "pressure": 1.3e7, "enth_mol": 18000}

    initialization_tester(m, dof=3, state_args=state_args)
    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)

    # check material balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].flow_mol
                - m.fs.unit.control_volume.properties_out[0].flow_mol
            ),
            abs=1e-3,
        )
        == 0
    )

    # pressure drop
    assert pytest.approx(-219382.990, rel=1e-3) == pyo.value(m.fs.unit.deltaP[0])
    # check energy balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].enth_mol
                - m.fs.unit.control_volume.properties_out[0].enth_mol
            ),
            abs=1e-3,
        )
        == 0
    )


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_pipe_vaporphase():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterPipe(
        dynamic=False,
        property_package=m.fs.properties,
        has_holdup=True,
        has_heat_transfer=False,
        has_pressure_change=True,
        water_phase="Vap",
        contraction_expansion_at_end="None",
    )
    m.fs.unit.diameter.fix(0.04)
    m.fs.unit.length.fix(40)
    m.fs.unit.number_of_pipes.fix(100)
    m.fs.unit.elevation_change.fix(25)
    m.fs.unit.fcorrection_dp.fix(1.0)

    state_args = {"flow_mol": 10000, "pressure": 3.6e6, "enth_mol": 53942}

    initialization_tester(m, dof=3, state_args=state_args)
    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)

    assert pyo.value(m.fs.unit.control_volume.properties_in[0].vapor_frac) == 1

    # check material balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].flow_mol
                - m.fs.unit.control_volume.properties_out[0].flow_mol
            ),
            abs=1e-3,
        )
        == 0
    )

    # pressure drop
    assert pytest.approx(-539105.588, rel=1e-3) == pyo.value(m.fs.unit.deltaP[0])
    # check energy balance
    assert (
        pytest.approx(
            pyo.value(
                m.fs.unit.control_volume.properties_in[0].enth_mol
                - m.fs.unit.control_volume.properties_out[0].enth_mol
            ),
            abs=1e-3,
        )
        == 0
    )
